@@ -1,9 +1,11 @@
+# -*- encoding: utf-8 -*-
 require 'sequel'
 require 'json'
 
 require_relative 'razor/version'
 require_relative 'razor/util'
 require_relative 'razor/middleware'
+require_relative 'razor/validation'
 require_relative 'razor/config'
 require_relative 'razor/matcher'
 require_relative 'razor/task'
@@ -13,3 +15,5 @@ require_relative 'razor/data'
 require_relative 'razor/view'
 require_relative 'razor/broker_type'
 require_relative 'razor/ipmi'
+require_relative 'razor/help'
+require_relative 'razor/command'
