@@ -1,3 +1,4 @@
+# -*- encoding: utf-8 -*-
 module Razor::Data
   class DuplicateNodeError < RuntimeError
     attr_reader :hw_info, :nodes
@@ -8,9 +9,10 @@ module Razor::Data
     end
 
     def message
-      nodes_message =
-        node_ids.map { |h| "(name=#{h[:name]}, id=#{h[:id]})" }.join(",")
-      "Multiple nodes match hw_info #{hw_info}. Nodes: #{nodes_message}"
+      # TRANSLATORS: the name of the node, and the hardware ID of it.
+      template = _("(name=%{name}, id=%{id})")
+      message  = node_ids.map { |h| template % {name: h[:name], id: h[:id]} }.join(", ")
+      _("Multiple nodes match hw_info %{hw_info}. Nodes: %{nodes}") % {hw_info: hw_info, nodes: message}
     end
 
     def log_to_nodes!
@@ -29,6 +31,13 @@ module Razor::Data
   end
 
   class Node < Sequel::Model
+    # Since we generate the name in the database using a trigger, the
+    # auto-validation plugin sets up the wrong constraint checks.
+    # This updates it to fix that, by moving from "must be present" to
+    # "validate not null if explicitly supplied", which allows the database to
+    # assign the default if it is not present.
+    auto_validate_not_null_columns.delete(:name)
+    auto_validate_explicit_not_null_columns << :name
 
     #See the method schema_type_class() for some special considerations
     #regarding the use of serialization.
@@ -37,6 +46,23 @@ module Razor::Data
 
     plugin :typecast_on_load, :hw_info
 
+    # The policy that governs how this node is to be or has been
+    # installed. Having policy set does not mean that the node is
+    # installed.  A node is only considered installed when the policy's
+    # task has successfully completed and stage_done('finished') has been
+    # called; once a node is installed we must consider what's on it
+    # precious to the user and will not try to alter the node without
+    # explicit user interaction, e.g. the user calling 'reinstall-node'.
+    #
+    # To be precise, here are the combinations of +policy+ and +installed+
+    # for a node and what they mean:
+    #
+    #   policy  | installed | meaning
+    #  -------------------------------------------------------------------
+    #   truthy  |  truthy   | successfully finished installation
+    #   truthy  |   nil     | in process of working through task
+    #     nil   |  truthy   | installed, but we don't know how it was done
+    #     nil   |   nil     | available for policy matching
     many_to_one :policy
     one_to_many :node_log_entries
 
@@ -54,11 +80,14 @@ module Razor::Data
       publish('eval_tags') if need_eval_tags
     end
 
-    # Return a 'name'; for now this is a fixed generated string
-    # @todo lutter 2013-08-30: figure out a way for users to control how
-    # node names are set
-    def name
-      id.nil? ? nil : "node#{id}"
+    def before_create
+      # Support users configuring that we mark all newly discovered nodes as
+      # "installed" already, despite having no policy.
+      if Razor.config['protect_new_nodes']
+        self.installed    = '+protected'
+        self.installed_at = Time.now
+      end
+      super
     end
 
     # Set the hardware info from a hash.
@@ -85,7 +114,7 @@ module Razor::Data
     def task
       if policy
         policy.task
-      elsif installed
+      elsif installed and registered?
         Razor::Task.noop_task
       else
         Razor::Task.mk_task
@@ -138,10 +167,27 @@ module Razor::Data
       # (otherwise we could have symbols, which will turn into strings on
       # reloading)
       entry = JSON::parse(entry.to_json)
-      TorqueBox::Logger.new.info("#{name}: #{entry.inspect}")
+
       add_node_log_entry(:entry => entry)
     end
 
+    # Return +true+ if the node has fully registered, i.e. has sent us its
+    # facts at least once. This supposes that the application makes sure
+    # that facts are initially, and whenever the need arises to change
+    # hw_info based on facts, set through +Node.register+
+    def registered?
+      ! facts.nil?
+    end
+
+    def installed=(value)
+      # @todo danielp 2014-04-15: Unfortunately, we can't store false into the
+      # database without also storing a time, but time should be nil if we are
+      # not installed.  Should fix that, one day, I suspect.
+      super(value ? value : nil)
+      self.installed_at = if value then DateTime.now else nil end
+      return value
+    end
+
     def bind(policy)
       self.policy = policy
       self.boot_count = 1
@@ -157,7 +203,6 @@ module Razor::Data
       #    require a flag to remember whether we've already booted into a
       #    policy or not
       self.installed = nil
-      self.installed_at = nil
       self.root_password = policy.root_password
       self.hostname = policy.hostname_pattern.gsub(/\$\{\s*id\s*\}/, id.to_s)
 
@@ -188,26 +233,27 @@ module Razor::Data
 
     def validate
       super
+
       unless hw_info.nil?
         # PGArray is not an Array, just behaves like one
         hw_info.is_a?(Sequel::Postgres::PGArray) or hw_info.is_a?(Array) or
-          errors.add(:hw_info, "must be an array")
+          errors.add(:hw_info, _("must be an array"))
         hw_info.each do |p|
           pair = p.split("=", 2)
           pair.size == 2 or
-            errors.add(:hw_info, "entry '#{p}' is not in the format 'key=value'")
+            errors.add(:hw_info, _("entry '%{raw}' is not in the format 'key=value'") % {raw: p})
           (pair[1].nil? or pair[1] == "") and
-            errors.add(:hw_info, "entry '#{p}' does not have a value")
-          Razor::Config::HW_INFO_KEYS.include?(pair[0]) or
-            errors.add(:hw_info, "entry '#{p}' uses an unknown key #{pair[0]}")
+            errors.add(:hw_info, _("entry '%{raw}' does not have a value") % {raw: p})
+          (Razor::Config::HW_INFO_KEYS.include?(pair[0]) or pair[0] =~ /^fact_/) or
+            errors.add(:hw_info, _("entry '%{raw}' uses an unknown key %{key}") % {raw: p, key: pair[0]})
           # @todo lutter 2013-09-03: we should do more checking, e.g. that
           # MAC addresses are sane
         end
       end
 
       if ipmi_hostname.nil?
-        ipmi_username and errors.add(:ipmi_username, 'you must also set an IPMI hostname')
-        ipmi_password and errors.add(:ipmi_password, 'you must also set an IPMI hostname')
+        ipmi_username and errors.add(:ipmi_username, _('you must also set an IPMI hostname'))
+        ipmi_password and errors.add(:ipmi_password, _('you must also set an IPMI hostname'))
       end
     end
 
@@ -234,14 +280,14 @@ module Razor::Data
       new_metadata = metadata
 
       if data['update']
-        data['update'].is_a? Hash or raise ArgumentError, 'update must be a hash'
+        data['update'].is_a? Hash or raise ArgumentError, _('update must be a hash')
         replace = (not [true, 'true'].include?(data['no_replace']))
         data['update'].each do |k,v|
           new_metadata[k] = v if replace or not new_metadata[k]
         end
       end
       if data['remove']
-        data['remove'].is_a? Array or raise ArgumentError, 'remove must be an array'
+        data['remove'].is_a? Array or raise ArgumentError, _('remove must be an array')
         data['remove'].each do |k,v|
           new_metadata.delete(k)
         end
@@ -272,20 +318,15 @@ module Razor::Data
       # time, i.e. have the update statement do 'last_checkin = now()' but
       # that is currently not possible with Sequel
       self.last_checkin = Time.now
-      action = :none
-      match_and_bind unless policy
+      match_and_bind unless (installed or policy)
       if policy
         log_append(:action => :reboot, :policy => policy.name)
-        action = :reboot
+      elsif installed
+        log_append(:action => :reboot, :task => 'noop',
+          :msg => _("Node has no policy but is installed. Booting locally"))
       end
       save_changes
-      { :action => action }
-    end
-
-    def self.find_by_name(name)
-      # We currently do not store the name in the DB; this just reverses
-      # what the +#name+ method does and looks up by id
-      self[$1] if name =~ /\Anode([0-9]+)\Z/
+      { :action => (installed or policy) ? :reboot : :none }
     end
 
     # Normalize the hardware info. Be very careful when you change this
@@ -295,25 +336,49 @@ module Razor::Data
     # Besides the keys coming in from the MK, +hw_info+ might also be a
     # +hw_hash+, implying that +mac+ might be an array of MAC addresses
     def self.canonicalize_hw_info(hw_info)
-      if macs = hw_info["mac"]
-        macs = [ macs ] unless macs.is_a? Array
-        macs = macs.map { |mac| mac.gsub(":", "-") }
-        # hw_info might contain an array of mac's; spread that out
-        hw_info = hw_info.to_a.reject! {|k,v| k == "mac" } +
-                  ["mac"].product(macs)
-      end
-      hw_info.map do |k,v|
+      hw_info = hw_info.dup
+
+      facts = (hw_info.delete("facts") || {}).map { |k,v| ["fact_#{k}",v] }
+
+      # Spread the (possible) array of macs out into a list of pairs
+      # [["mac", ..], ["mac", ..]]
+      macs = Array(hw_info.delete("mac")).map { |mac| mac.gsub(":", "-") }
+      macs = ["mac"].product(macs)
+
+      (hw_info.to_a + facts + macs).reject do |_,v|
+        v.nil? || v.strip == ""
+      end.map do |k,v|
         # We treat the netXXX keys special so that our hw_info is
         # independent of the order in which the BIOS enumerates NICs. We
         # also don't care about case
         k = "mac" if k =~ /net[0-9]+/
         [k.downcase, v.strip.downcase]
-      end.select do |k, v|
-        Razor::Config::HW_INFO_KEYS.include?(k) && v && v != ""
+      end.select do |k, _|
+        Razor::Config::HW_INFO_KEYS.include?(k) || k.start_with?('fact_')
       end.sort do |a, b|
         # Sort the [key, value] pairs lexicographically
         a[0] == b[0] ? a[1] <=> b[1] : a[0] <=> b[0]
-      end.map { |pair| "#{pair[0]}=#{pair[1]}" }
+      end.map do |k,v|
+        "#{k}=#{v}"
+      end
+    end
+
+    # Find all nodes matching the hardware criteria in +params+ which must
+    # be in a format that +canonicalize_hw_info+ understands. Return a
+    # pair, where the first part is an array of nodes, and the second part
+    # the +hw_info+ that was used for the lookup
+    def self.find_by_hw_info(params)
+      # For matching nodes, we only consider the +hw_info+ values named in
+      # the 'match_nodes_on' config setting.
+      hw_match = canonicalize_hw_info(params).select do |p|
+        name = p.split("=")[0]
+        Razor.config['match_nodes_on'].include?(name) or
+          name.start_with?('fact_')
+      end
+
+      hw_match.empty? and raise ArgumentError, _("Lookup was given %{keys}, none of which are configured as match criteria in match_nodes_on (%{match_nodes_on})") % {keys: params.keys, match_nodes_on: Razor.config['match_nodes_on']}
+
+      [self.where(:hw_info.pg_array.overlaps(hw_match)).all, hw_match]
     end
 
     # Look up a node by its hw_info; any node whose hw_info overlaps with
@@ -324,26 +389,20 @@ module Razor::Data
       dhcp_mac = params.delete("dhcp_mac")
       dhcp_mac = nil if !dhcp_mac.nil? and dhcp_mac.empty?
 
-      hw_info = canonicalize_hw_info(params)
-      # For matching nodes, we only consider the +hw_info+ values named in
-      # the 'match_nodes_on' config setting
-      hw_match = hw_info.select do |p|
-        Razor.config['match_nodes_on'].include?(p.split("=")[0])
-      end
-      hw_match.empty? and raise ArgumentError, "Lookup was given #{params.keys}, none of which are configured as match criteria in match_nodes_on (#{Razor.config['match_nodes_on']})"
-      nodes = self.where(:hw_info.pg_array.overlaps(hw_match)).all
+      nodes, hw_info = self.find_by_hw_info(params)
       if nodes.size == 0
         self.create(:hw_info => hw_info, :dhcp_mac => dhcp_mac)
       elsif nodes.size == 1
         node = nodes.first
+        # We do not want to update the hw_info at this point; all we know
+        # is what iPXE sees and that might not be the complete picture. If
+        # we fail to identify an existing node because of HW changes, we
+        # rely on the fact that a new node gets created, and we later on
+        # call register to merge the existing and the new node
         unless dhcp_mac.nil? || node.dhcp_mac == dhcp_mac
           node.dhcp_mac = dhcp_mac
           node.save
         end
-        if hw_info != node.hw_info
-          node.hw_info = hw_info
-          node.save
-        end
         node
       else
         # We have more than one node matching hw_info; fail
@@ -351,6 +410,57 @@ module Razor::Data
       end
     end
 
+    # Use the facts +facts+ to fully register the node; this is a
+    # counterpart to +lookup+; while +lookup+ uses the much more restricted
+    # information provided by iPXE, this method relies on all the facts
+    # we've gathered and can therefore use much more thorough information
+    # about the node to set +hw_info+
+    #
+    # As an example, iPXE generally (because we use undionly.kpxe) will
+    # only see one NIC, whereas we have all of them in our facts
+    def self.register(facts)
+      macs = facts.select { |k,_| k.start_with?('macaddress') }.
+        map { |_, v| v }.uniq
+
+      # @todo lutter 2014-05-19: we should also fill the asset tag; I am
+      # not sure which fact coresponds to the asset tag
+      nodes, hw_info = self.find_by_hw_info({
+        'mac'    => macs,
+        'serial' => facts['serialnumber'],
+        'uuid'   => facts['uuid'],
+        'facts'  => facts.select { |k, _| Razor.config.fact_match_on?(k) }
+      })
+
+      if nodes.size == 0
+        # Should not happen
+        Razor.logger.error("Failed to find node with facts #{facts.inspect} for registration")
+        return nil
+      else
+        # If all but one of the nodes have not been registered yet, we can
+        # merge them into the registered node. Otherwise, we complain
+        if nodes.any? { |n| n.registered? }
+          keep_node  = nodes.find { |n| n.registered? }
+          kill_nodes = nodes.reject { |n| n.registered? }
+          if kill_nodes.size != nodes.size - 1
+            # We found more than one node that was already registered
+            raise DuplicateNodeError.new(hw_info, nodes)
+          end
+        else
+          keep_node  = nodes.first
+          kill_nodes = nodes[1..-1]
+        end
+
+        keep_node.hw_info = hw_info
+
+        kill_nodes.each do |kill_node|
+          kill_node.node_log_entries_dataset.update(:node_id => keep_node.id)
+          kill_node.destroy
+        end
+        keep_node.save
+        keep_node
+      end
+    end
+
     def self.stage_done(node_id, name = "")
       node = self[node_id]
       name = node.boot_count if name.nil? or name.empty?
@@ -358,7 +468,6 @@ module Razor::Data
       node.boot_count += 1
       if name == "finished" and node.policy
         node.installed = node.policy.name
-        node.installed_at = DateTime.now
       end
       node.save
     end
@@ -425,11 +534,10 @@ module Razor::Data
       end
     end
 
-    # Request a reboot from the machine via IPMI.  This supports both hard and
-    # soft reboots.  This is synchronous, and is expected to be called in the
-    # background from the message queue.
-    def reboot!(hard)
-      Razor::IPMI.reset(self, hard)
+    # Request a reboot from the machine via IPMI.  This is synchronous, and is
+    # expected to be called in the background from the message queue.
+    def reboot!
+      Razor::IPMI.reset(self)
     end
 
     # Turn the node on.
